import unittest

import numpy as np

import fwdpy11


class GenerationRecorder(object):
    def __init__(self):
        self.generations = []

    def __call__(self, pop):
        self.generations.append(pop.generation)


class testWFevolve(unittest.TestCase):
    @classmethod
    def setUpClass(self):

        self.pop = fwdpy11.DiploidPopulation(1000)
        self.rng = fwdpy11.GSLrng(42)
        self.recorder = GenerationRecorder()
        self.p = {
            "rates": (1e-3, 1e-3, 1e-3),
            "popsizes": np.array([1000] * 100, dtype=np.uint32),
            "simlen": 100,
            "nregions": [fwdpy11.Region(0, 1, 1)],
            "sregions": [fwdpy11.ExpS(0, 1, 1, -1e-2)],
            "recregions": [fwdpy11.Region(0, 1, 1)],
            "gvalue": fwdpy11.Multiplicative(2.0),
        }

    def testEvolve(self):

        params = fwdpy11.ModelParams(**self.p)
        fwdpy11.evolve_genomes(self.rng, self.pop, params, self.recorder)
        self.assertEqual(self.recorder.generations, [i + 1 for i in range(100)])
        self.p["popsizes"] = np.array([self.pop.N] * 24, dtype=np.uint32)
        self.p["simlen"] = len(self.p["popsizes"])
        params = fwdpy11.ModelParams(**self.p)
        fwdpy11.evolve_genomes(self.rng, self.pop, params, self.recorder)
        self.assertEqual(self.recorder.generations, [i + 1 for i in range(124)])


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing

import attr
import numpy as np

import fwdpy11

from .class_decorators import (attr_add_asblack, attr_class_pickle_with_super,
                               attr_class_to_from_dict,
                               attr_class_to_from_dict_no_recurse)

_common_attr_attribs = {"frozen": True, "auto_attribs": True, "repr_ns": "fwdpy11"}


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(**_common_attr_attribs)
class MassMigration(fwdpy11._fwdpy11._ll_MassMigration):
    """
    Mass migration events.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param when: When the mass migration happens
    :type when: int
    :param source: The source deme for individuals to move
    :type source: int
    :param destination: The deme to where individuals we be moved
    :type destination: int
    :param fraction: The fraction of `source` to move to `destination`.
    :type fraction: float
    :param move_individuals: If ``True``, the event moves individuals.
                             If ``False``, individuals are copied.
    :type move_individuals: bool
    :param resets_growth_rate: (True) Whether or not to reset the
                               growth rates of `source` and `destination` to
                               :data:`fwdpy11.NOGROWTH`
    :type resets_growth_rate: bool

    .. note::

        It is a bit simpler to construct instances of this class
        using :func:`fwdpy11.move_individuals` or :func:`fwdpy11.copy_individuals`.

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    when: int = attr.ib()
    source: int = attr.ib()
    destination: int = attr.ib()
    fraction: float = attr.ib()
    move_individuals: bool = attr.ib(default=True)
    resets_growth_rate: bool = attr.ib(default=True)

    def __attrs_post_init__(self):
        super(MassMigration, self).__init__(
            self.when,
            self.source,
            self.destination,
            self.fraction,
            self.move_individuals,
            self.resets_growth_rate,
        )


def move_individuals(
    when: int,
    source: int,
    destination: int,
    fraction: float,
    resets_growth_rate: bool = True,
) -> MassMigration:
    """
    :param when: The generation when the event occurs
    :type when: int
    :param source: The source deme for individuals to move
    :type source: int
    :param destination: The deme to where individuals we be moved
    :type destination: int
    :param fraction: The fraction of `source` to move to `destination`.
    :type fraction: float
    :param resets_growth_rate: (True) Whether or not to reset the
                               growth rates of `source` and `destination` to
                               :data:`fwdpy11.NOGROWTH`
    :type resets_growth_rate: bool

    :rtype: :class:`fwdpy11.MassMigration`
    """
    return MassMigration(when, source, destination, fraction, True, resets_growth_rate)


def copy_individuals(
    when: int,
    source: int,
    destination: int,
    fraction: float,
    resets_growth_rate: bool = True,
) -> MassMigration:
    """
    :param when: The generation when the event occurs
    :type when: int
    :param source: The source deme for individuals to copy
    :type source: int
    :param destination: The deme to where individuals we be copied
    :type destination: int
    :param fraction: The fraction of `source` to copy to `destination`.
    :type fraction: float
    :param resets_growth_rate: (True) Whether or not to reset the
                               growth rates of `source` and `destination` to
                               :data:`fwdpy11.NOGROWTH`
    :type resets_growth_rate: bool

    :rtype: :class:`fwdpy11.MassMigration`
    """
    return MassMigration(when, source, destination, fraction, False, resets_growth_rate)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(**_common_attr_attribs)
class SetDemeSize(fwdpy11._fwdpy11._ll_SetDemeSize):
    """
    Set the size of a deme at a given time.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param when: The generation when the event occurs
    :type when: int
    :param deme: The deme whose size will change
    :type deme: int
    :param new_size: The new size
    :type new_size: int
    :param resets_growth_rate: (True) If deme size change resets
                               growth rate to :data:`fwdpy11.NOGROWTH`
    :type resets_growth_rate: bool

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    when: int = attr.ib()
    deme: int = attr.ib()
    new_size: int = attr.ib()
    resets_growth_rate: bool = attr.ib(default=True)

    def __attrs_post_init__(self):
        super(SetDemeSize, self).__init__(
            self.when, self.deme, self.new_size, self.resets_growth_rate
        )


@attr_class_to_from_dict
@attr_add_asblack
@attr_class_pickle_with_super
@attr.s(**_common_attr_attribs)
class SetExponentialGrowth(fwdpy11._fwdpy11._ll_SetExponentialGrowth):
    """
    Set the growth rate of a deme at a given time.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param when: The generation when the event occurs
    :type when: int
    :param deme: The deme whose growth rate will change
    :type deme: int
    :param G: The new growth rate
    :type G: float

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    when: int = attr.ib()
    deme: int = attr.ib()
    G: float = attr.ib()

    def __attrs_post_init__(self):
        super(SetExponentialGrowth, self).__init__(self.when, self.deme, self.G)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(**_common_attr_attribs)
class SetSelfingRate(fwdpy11._fwdpy11._ll_SetSelfingRate):
    """
    Set the selfing probability within a deme at a given time.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param when: The generation when the event occurs
    :type when: int
    :param deme: The deme whose selfing probability will change
    :type deme: int
    :param S: The new selfing probability
    :type S: float

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    when: int = attr.ib()
    deme: int = attr.ib()
    S: float = attr.ib()

    def __attrs_post_init__(self):
        super(SetSelfingRate, self).__init__(self.when, self.deme, self.S)

    def __getstate__(self):
        return self.asdict()

    def __setstate__(self, d):
        self.__dict__.update(d)
        super(SetSelfingRate, self).__init__(**d)


@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(eq=False, **_common_attr_attribs)
class MigrationMatrix(fwdpy11._fwdpy11._ll_MigrationMatrix):
    """
    The forward migration matrix for a simulation.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param migmatrix: A square matrix of non-negative floats.
    :type migmatrix: numpy.ndarray
    :param scaled: (True) If entries in `migmatrix` will be
                   multiplied by deme sizes during simulation
    :type scaled: bool

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    migmatrix: np.ndarray = attr.ib()
    scaled: bool = attr.ib(default=False)

    def __attrs_post_init__(self):
        super(MigrationMatrix, self).__init__(self.migmatrix, self.scaled)

    def __eq__(self, other):
        return self.scaled == other.scaled and np.array_equal(
            self.migmatrix, other.migmatrix
        )

    @property
    def M(self):
        return self.migmatrix

    @property
    def shape(self):
        return self.migmatrix.shape


def _set_migration_rates_convert_deme(i: typing.Optional[int]) -> int:
    if i is None:
        return -1
    return int(i)


@attr_add_asblack
@attr_class_to_from_dict
@attr.s(eq=False, **_common_attr_attribs)
class SetMigrationRates(fwdpy11._fwdpy11._ll_SetMigrationRates):
    """
    Set the migration parameters of a simulation at a given time.
    May be used to set either the migration rates from a given
    deme or the entire migration matrix.

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param when: The generation when the event occurs
    :type when: int
    :param deme: The row index of the migration matrix
    :type deme: int or None
    :param migrates: The migration rates from `deme` to all other populations.
    :type migrates: list or numpy.ndarray

    In order to change the entire migration matrix, pass ``None`` (or -1)
    to ``deme`` and a 2d ``numpy.ndarray`` for ``migrates``.  A value of
    ``None`` will be converted to ``-1`` in this case.

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    when: int = attr.ib()
    deme: typing.Optional[int] = attr.ib(converter=_set_migration_rates_convert_deme)
    migrates: np.ndarray = attr.ib()

    def __attrs_post_init__(self):
        if self.deme >= 0:
            try:
                super(SetMigrationRates, self).__init__(
                    self.when, self.deme, self.migrates.tolist()
                )
            except AttributeError:
                super(SetMigrationRates, self).__init__(
                    self.when, self.deme, self.migrates
                )
        else:
            super(SetMigrationRates, self).__init__(self.when, self.migrates.tolist())

    def __getstate__(self):
        return self.asdict()

    def __setstate__(self, d):
        self.__dict__.update(d)
        if self.deme == -1:
            super(SetMigrationRates, self).__init__(self.when, self.migrates)
        else:
            super(SetMigrationRates, self).__init__(self.when, self.deme, self.migrates)

    def __eq__(self, other):
        return (
            self.deme == other.deme
            and self.when == other.when
            and np.array_equal(self.migrates, other.migrates)
        )


def _convert_demographic_events_list(
    o: typing.Optional[typing.List],
) -> typing.Optional[typing.List]:
    if o is None:
        return o

    try:
        rv = sorted(o, key=lambda x: (x.when, x.source, x.move_individuals))
    except AttributeError:
        rv = sorted(o, key=lambda x: (x.when, x.deme))

    return rv


def _convert_set_deme_sizes(
    o: typing.Optional[np.ndarray],
) -> typing.Optional[np.ndarray]:
    if o is None:
        return o
    if isinstance(o, np.ndarray):
        N = o[0]
        rv = [SetDemeSize(when=0, deme=0, new_size=N, resets_growth_rate=True)]
        for i in range(1, len(o)):
            if o[i] != N:
                N = o[i]
                rv.append(
                    SetDemeSize(when=i, deme=0, new_size=N, resets_growth_rate=True)
                )

        return rv

    return _convert_demographic_events_list(o)


def _convert_migmatrix(o):
    if o is None:
        return o
    try:
        return MigrationMatrix(o.migmatrix, o.scaled)
    except AttributeError:
        pass
    if isinstance(o, tuple):
        return MigrationMatrix(*o)
    return MigrationMatrix(o)


@attr_add_asblack
@attr_class_to_from_dict_no_recurse
@attr.s(**_common_attr_attribs)
class DiscreteDemography(fwdpy11._fwdpy11._ll_DiscreteDemography):
    """
    Representation of demographic events acting on
    discrete demes (sub-populations).

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param mass_migrations: Instances of :class:`fwdpy11.MassMigration`
    :type mass_migrations: None or list[fwdpy11.MassMigration]
    :param set_growth_rates: Instances of :class:`fwdpy11.SetExponentialGrowth`
    :type set_growth_rates: None or list[fwdpy11.SetExponentialGrowth]
    :param set_deme_sizes: Instances of :class:`fwdpy11.SetDemeSize`
    :type set_deme_sizes: None or list[fwdpy11.SetDemeSize]
    :param set_selfing_rates: Instances of :class:`fwdpy11.SetSelfingRate`
    :type set_selfing_rates: None or list[fwdpy11.SetSelfingRate]
    :param migmatrix: A migraton matrix. See :ref:`migration`. If not
                      ``None``, then input are converted into an
                      instance of :class:`fwdpy11.MigrationMatrix`.
    :type migmatrix: None or numpy.ndarray or fwdpy11.MigrationMatrix
    :param set_migration_rates: Instances of :class:`fwdpy11.SetMigrationRates`
    :type set_migration_rates: None or list[fwdpy11.SetMigrationRates]

    .. versionadded:: 0.5.3

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    mass_migrations: typing.Optional[typing.List[MassMigration]] = attr.ib(
        default=None, converter=_convert_demographic_events_list
    )
    set_growth_rates: typing.Optional[typing.List[SetExponentialGrowth]] = attr.ib(
        default=None, converter=_convert_demographic_events_list
    )
    set_deme_sizes: typing.Optional[typing.List[SetDemeSize]] = attr.ib(
        default=None, converter=_convert_set_deme_sizes
    )
    set_selfing_rates: typing.Optional[typing.List[SetSelfingRate]] = attr.ib(
        default=None, converter=_convert_demographic_events_list
    )
    migmatrix: typing.Optional[typing.Union[np.ndarray, MigrationMatrix]] = attr.ib(
        default=None, converter=_convert_migmatrix
    )
    set_migration_rates: typing.Optional[typing.List[SetMigrationRates]] = attr.ib(
        default=None, converter=_convert_demographic_events_list
    )

    def __attrs_post_init__(self):
        super(DiscreteDemography, self).__init__(
            mass_migrations=self.mass_migrations,
            set_growth_rates=self.set_growth_rates,
            set_deme_sizes=self.set_deme_sizes,
            set_selfing_rates=self.set_selfing_rates,
            migmatrix=self.migmatrix,
            set_migration_rates=self.set_migration_rates,
        )

    def __getstate__(self):
        return (self.asdict(), self._state_asdict())

    def __setstate__(self, t):
        self.__dict__.update(t[0])
        super(DiscreteDemography, self).__init__(**t[0])
        self._reset_state(t[1])

    def _timed_events(self):
        for i in [
            self.mass_migrations,
            self.set_deme_sizes,
            self.set_growth_rates,
            self.set_selfing_rates,
            self.set_migration_rates,
        ]:
            yield i



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import pickle
import unittest

import numpy as np

import fwdpy11


def _starting_pdict():
    return {
        "nregions": [],
        "sregions": [fwdpy11.ExpS(0, 1, 1, -0.1, 0.25)],
        "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
        "rates": [0, 0.01, None],
        "demography": fwdpy11.DiscreteDemography(),
        "simlen": 10,
        "gvalue": fwdpy11.Multiplicative(
            2.0, fwdpy11.GSS(fwdpy11.Optimum(VS=1.0, optimum=0.0))
        ),
    }


class TestPickling(unittest.TestCase):
    """
    Test pickling a couple of models that use all of the features.

    Note: test is limited as not all parameter types support __eq__ (yet)
    """

    @classmethod
    def setUp(self):
        self.pdict = _starting_pdict()

    def test_pickle_IM_with_selfing(self):
        import fwdpy11.demographic_models.IM as IM

        dmodel = IM.two_deme_IM(1000, 100, 0.25, [0.5, 3], [0.1, 0.25])
        temp = dmodel.model.asdict()
        temp["set_selfing_rates"] = [fwdpy11.SetSelfingRate(0, 0, 0.1)]
        im = fwdpy11.DiscreteDemography(**temp)
        self.pdict["demography"] = im
        params = fwdpy11.ModelParams(**self.pdict)
        pi = pickle.dumps(params)
        up = pickle.loads(pi)
        self.assertEqual(params.demography, up.demography)

    def test_pickle_IM_with_whole_migmatrix_reset(self):
        import fwdpy11.demographic_models.IM as IM

        dmodel = IM.two_deme_IM(1000, 100, 0.25, [0.5, 3], [0.1, 0.25])
        temp = dmodel.model.asdict()
        temp["set_migration_rates"].append(
            fwdpy11.SetMigrationRates(10000 + 10, None, np.eye(2))
        )
        im = fwdpy11.DiscreteDemography(**temp)
        self.pdict["demography"] = im
        params = fwdpy11.ModelParams(**self.pdict)
        pi = pickle.dumps(params)
        up = pickle.loads(pi)
        self.assertEqual(params.demography, up.demography)

    def test_pickle_tennessen(self):
        import fwdpy11.demographic_models.human as human

        tennessen = human.tennessen()
        self.pdict["demography"] = tennessen
        params = fwdpy11.ModelParams(**self.pdict)
        pi = pickle.dumps(params)
        up = pickle.loads(pi)
        self.assertEqual(params.demography, up.demography)

    def test_pickle_model_with_scaled_migration_matrix(self):
        m = fwdpy11.DiscreteDemography(migmatrix=(np.eye(2), True))
        self.pdict["demography"] = m
        params = fwdpy11.ModelParams(**self.pdict)
        pi = pickle.dumps(params)
        up = pickle.loads(pi)
        self.assertEqual(params.demography, up.demography)


class TestEval(unittest.TestCase):
    """
    Test round-tripping model parameters
    through strings.

    These tests reveal quirks, such
    as not getting the name spaces
    of other packages such as numpy.
    """

    def test_round_trip_through_str(self):
        import fwdpy11.demographic_models.human as human
        from numpy import array  # NOQA

        tennessen = human.tennessen()
        pdict = _starting_pdict()
        pdict["demography"] = tennessen
        params = fwdpy11.ModelParams(**pdict)
        params_s = str(params)
        params_eval = eval(params_s)
        self.assertEqual(params, params_eval)

    def test_round_trip_through_bytes(self):
        import fwdpy11.demographic_models.human as human
        from numpy import array  # NOQA

        tennessen = human.tennessen()
        pdict = _starting_pdict()
        pdict["demography"] = tennessen
        params = fwdpy11.ModelParams(**pdict)
        params_s = bytes(str(params).encode("utf-8"))
        params_eval = eval(params_s.decode("utf-8"))
        self.assertEqual(params, params_eval)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import json

import numpy as np
import tskit

import fwdpy11.tskit_tools


def _alive_nodes(self):
    """
    List of alive nodes corresponding to individuals.

    :rtype: numpy.ndarray

    .. vesionadded:: 0.5.3
    """
    md = np.array(self.diploid_metadata, copy=False)
    return md["nodes"].flatten()


def _get_times(self):
    amd = np.array(self.ancient_sample_metadata, copy=False)
    nodes = np.array(self.tables.nodes, copy=False)
    times = nodes["time"][amd["nodes"][:, 0]]
    utimes = np.unique(times)
    return times, utimes


def _traverse_sample_timepoints(self, include_alive=True):
    """
    Return an iterator over all sample time points.
    The iterator yields time, nodes, and metadata.

    :param include_alive: If True, include currently-alive individuals.

    .. versionadded :: 0.5.1

    .. versionchanged:: 0.5.3

       Monkey-patched into pybind11 class
    """
    amd = np.array(self.ancient_sample_metadata, copy=False)
    amd.flags.writeable = False
    times, utimes = _get_times(self)
    for uti in utimes:
        mdidx = np.where(times == uti)[0]
        sample_nodes = amd["nodes"][mdidx].flatten()
        sample_nodes.flags.writeable = False
        mdslice = amd[mdidx][:]
        mdslice.flags.writeable = False
        yield uti, sample_nodes, mdslice

    if include_alive is True:
        md = np.array(self.diploid_metadata, copy=False)
        nodes = md["nodes"].flatten()
        md.flags.writeable = False
        nodes.flags.writeable = False
        yield self.generation, nodes, md


# NOTE: mutation origin times are recorded forwards in time,
# So we convert them into mutation ages.
def _generate_mutation_metadata(self):
    muts = []
    for mr in self.tables.mutations:
        m = self.mutations[mr.key]
        d = {
            "s": m.s,
            "h": m.h,
            "age": self.generation - m.g + 1,
            "label": m.label,
            "esizes": list(m.esizes),
            "heffects": list(m.heffects),
            "neutral": m.neutral,
            "key": mr.key,
        }
        muts.append(str(d).encode("utf-8"))
    return tskit.pack_bytes(muts)


def _initializePopulationTable(node_view, tc):
    population_metadata = []
    for i in sorted(np.unique(node_view["deme"])):
        md = "deme" + str(i)
        population_metadata.append(md.encode("utf-8"))

    pmd, pmdo = tskit.pack_bytes(population_metadata)
    tc.populations.set_columns(metadata=pmd, metadata_offset=pmdo)


def _generate_individual_metadata(dmd, tc):
    strings = []
    for i in dmd:
        d = {
            "g": i.g,
            "e": i.e,
            "w": i.w,
            "geography": i.geography,
            "parents": i.parents,
            "sex": i.sex,
            "deme": i.deme,
            "label": i.label,
        }
        strings.append(str(d).encode("utf-8"))
    return strings


def _initializeIndividualTable(self, tc):
    """
    Returns node ID -> individual map
    """
    # First, alive individuals:
    individal_nodes = {}
    flags = []
    for i in range(self.N):
        individal_nodes[2 * i] = i
        individal_nodes[2 * i + 1] = i
        flags.append(fwdpy11.tskit_tools.INDIVIDUAL_IS_ALIVE)
    metadata_strings = _generate_individual_metadata(self.diploid_metadata, tc)

    # Now, preserved nodes
    num_ind_nodes = self.N
    for i in self.ancient_sample_metadata:
        assert i not in individal_nodes, "indivudal record error"
        individal_nodes[i.nodes[0]] = num_ind_nodes
        individal_nodes[i.nodes[1]] = num_ind_nodes
        num_ind_nodes += 1
        flag = fwdpy11.tskit_tools.INDIVIDUAL_IS_PRESERVED
        if (
            self.tables.nodes[i.nodes[0]].time == 0.0
            and self.tables.nodes[i.nodes[1]].time == 0.0
        ):
            flag |= fwdpy11.tskit_tools.INDIVIDUAL_IS_FIRST_GENERATION
        flags.append(flag)

    metadata_strings.extend(
        _generate_individual_metadata(self.ancient_sample_metadata, tc)
    )

    md, mdo = tskit.pack_bytes(metadata_strings)
    # flags = [0 for i in range(self.N + len(self.ancient_sample_metadata))]
    tc.individuals.set_columns(flags=flags, metadata=md, metadata_offset=mdo)
    return individal_nodes


def _dump_tables_to_tskit(self, parameters=None):
    """
    Dump the population's TableCollection into
    an tskit TreeSequence

    :param parameters: The simulation parameters for the provenance table.
    :type parameters: None or dict

    :rtype: tskit.TreeSequence

    .. versionchanged:: 0.8.2

        Added `parameters`.
        Generate provenance information for return value.
        The provenance information is validated using
        :func:`tskit.validate_provenance`, which may
        raise an exception.
    """
    from fwdpy11 import pybind11_version, gsl_version

    environment = tskit.provenance.get_environment(
        extra_libs={
            "gsl": {"version": gsl_version()["gsl_version"]},
            "pybind11": {"version": pybind11_version()["pybind11_version"]},
        }
    )

    provenance = {
        "schema_version": "1.0.0",
        "software": {"name": "fwdpy11", "version": f"{fwdpy11.__version__}"},
        "environment": environment,
        "parameters": {},
    }

    if parameters is not None:
        provenance["parameters"] = parameters

    tskit.validate_provenance(provenance)

    node_view = np.array(self.tables.nodes, copy=True)
    node_view["time"] -= node_view["time"].max()
    node_view["time"][np.where(node_view["time"] != 0.0)[0]] *= -1.0
    edge_view = np.array(self.tables.edges, copy=False)
    mut_view = np.array(self.tables.mutations, copy=False)

    tc = tskit.TableCollection(self.tables.genome_length)

    # We must initialize population and individual
    # tables before we can do anything else.
    # Attempting to set population to anything
    # other than -1 in an tskit.NodeTable will
    # raise an exception if the PopulationTable
    # isn't set up.
    _initializePopulationTable(node_view, tc)
    node_to_individual = _initializeIndividualTable(self, tc)

    individual = [-1 for i in range(len(node_view))]
    for k, v in node_to_individual.items():
        individual[k] = v
    flags = [1] * 2 * self.N + [0] * (len(node_view) - 2 * self.N)
    # Bug fixed in 0.3.1: add preserved nodes to samples list
    for i in self.tables.preserved_nodes:
        flags[i] = 1
    tc.nodes.set_columns(
        flags=flags,
        time=node_view["time"],
        population=node_view["deme"],
        individual=individual,
    )
    tc.edges.set_columns(
        left=edge_view["left"],
        right=edge_view["right"],
        parent=edge_view["parent"],
        child=edge_view["child"],
    )

    mpos = np.array([self.mutations[i].pos for i in mut_view["key"]])
    ancestral_state = np.zeros(len(mut_view), dtype=np.int8) + ord("0")
    ancestral_state_offset = np.arange(len(mut_view) + 1, dtype=np.uint32)
    tc.sites.set_columns(
        position=mpos,
        ancestral_state=ancestral_state,
        ancestral_state_offset=ancestral_state_offset,
    )

    derived_state = np.zeros(len(mut_view), dtype=np.int8) + ord("1")
    md, mdo = _generate_mutation_metadata(self)
    tc.mutations.set_columns(
        site=np.arange(len(mpos), dtype=np.int32),
        node=mut_view["node"],
        derived_state=derived_state,
        derived_state_offset=ancestral_state_offset,
        metadata=md,
        metadata_offset=mdo,
    )
    tc.provenances.add_row(json.dumps(provenance))
    return tc.tree_sequence()


def _deme_sizes(self, as_dict=False):
    """
    Return the number of individuals in each deme.

    :param as_dict: If True, return results as a dict.
                    If False, numpy arrays are returned

    The default behavior of this function is equivalent to:

    .. code-block:: python

        md = numpy.array(self.diploid_metadata, copy=False)
        return numpy.unique(md['deme'], return_counts=True)

    .. versionadded:: 0.6.0
    """
    md = np.array(self.diploid_metadata, copy=False)
    deme_sizes = np.unique(md["deme"], return_counts=True)
    if as_dict is False:
        return deme_sizes
    return {i: j for i, j in zip(deme_sizes[0], deme_sizes[1])}


def _patch_diploid_population(d):
    d.alive_nodes = property(_alive_nodes)
    d.sample_timepoints = _traverse_sample_timepoints
    d.dump_tables_to_tskit = _dump_tables_to_tskit
    d.deme_sizes = _deme_sizes



import precapitate
import recapitate
%load_ext line_profiler
%lprun -f recapitate.run_sim recapitate.run_sim(1000, 1000, 1000, 1000, "dtwf", 100, 101, 666)
%lprun -f precapitate.run_sim precapitate.run_sim(1000, 1000, 1000, 1000, "dtwf", 100, 101, 666)



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import argparse
import concurrent.futures
import sys
from collections import namedtuple

import msprime
import numpy as np
import pandas as pd

import fwdpy11

SimOutput = namedtuple("SimOutput", ["S2N", "Pi2N", "Sn", "Pin"])


def make_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument("-N", type=int, help="Diploid population size")
    parser.add_argument("--rho", type=float, help="4Nr")
    parser.add_argument("--theta", type=float, help="4Nu")
    parser.add_argument(
        "--simlen", type=int, help="Generations to run the forward simulation"
    )
    parser.add_argument("--seed", type=int, help="Random number seed")
    parser.add_argument("--nreps", type=int, help="Number of simulation replicates")
    parser.add_argument("--nsam", type=int, help="Sample size to analyze")

    optional = parser.add_argument_group("Optional arguments")
    optional.add_argument("--model", type=str, help="msprime model", default="dtwf")
    return parser


def run_msprime(Ne, rho, theta, model, seed, input_ts=None):
    if input_ts is None:
        ts = msprime.simulate(
            2 * Ne,
            Ne=Ne,
            model=model,
            random_seed=seed,
            recombination_rate=rho / Ne / 4,
            mutation_rate=theta / Ne / 4,
        )
    else:
        ts = msprime.simulate(
            from_ts=input_ts,
            Ne=Ne,
            model=model,
            random_seed=seed,
            recombination_rate=rho / Ne / 4,
        )

    return ts


def run_msprime_get_stats(Ne, nsam, rho, theta, model, seed):
    ts = run_msprime(Ne, rho, theta, model, seed)
    samples = [i for i in ts.samples()]
    S1 = len(ts.tables.sites)
    pi1 = ts.diversity([samples])[0]
    rsample = np.random.choice(samples, nsam, replace=False)
    S2 = ts.segregating_sites([rsample])[0]
    pi2 = ts.diversity([rsample])[0]
    return SimOutput(S1, pi1, S2, pi2)


def run_sim(N, rho, theta, simlen, model, nsam, fwdpy11_seed, msprime_seed):
    np.random.seed(msprime_seed)
    pop = fwdpy11.DiploidPopulation(N, 1.0)

    pdict = {
        "nregions": [],
        "sregions": [],
        "recregions": [
            fwdpy11.PoissonInterval(0, pop.tables.genome_length, rho / pop.N / 4)
        ],
        "rates": (0, 0, None),
        "gvalue": fwdpy11.Multiplicative(1),
        "demography": fwdpy11.DiscreteDemography(),
        "simlen": simlen,
    }
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(fwdpy11_seed)
    fwdpy11.evolvets(rng, pop, params, 100, preserve_first_generation=True)
    ts = pop.dump_tables_to_tskit()
    completed_ts = run_msprime(N, rho, 0.0, model, msprime_seed, ts)
    samples = pop.alive_nodes
    completed_ts = msprime.mutate(
        completed_ts, rate=theta / N / 4, random_seed=msprime_seed
    )
    S2N = completed_ts.segregating_sites([samples])[0]
    pi2N = completed_ts.diversity([samples])[0]
    rsample = np.random.choice(samples, nsam, replace=False)
    Sn = completed_ts.segregating_sites([rsample])[0]
    pin = completed_ts.diversity([rsample])[0]

    return SimOutput(S2N, pi2N, Sn, pin)


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args(sys.argv[1:])

    np.random.seed(args.seed)

    msprime_seeds = []
    fwdpy11_seeds = []

    for i in range(args.nreps):
        candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        while candidate in msprime_seeds:
            candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        msprime_seeds.append(candidate)
        candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        while candidate in fwdpy11_seeds:
            candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        fwdpy11_seeds.append(candidate)

    results = []
    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = {
            executor.submit(
                run_sim,
                args.N,
                args.rho,
                args.theta,
                args.simlen,
                args.model,
                args.nsam,
                i,
                j,
            )
            for i, j in zip(msprime_seeds, fwdpy11_seeds)
        }
        for future in concurrent.futures.as_completed(futures):
            results.append(future.result())

    results = pd.DataFrame(results, columns=SimOutput._fields)
    results["watterson_n"] = results.Sn / (1.0 / (np.arange(args.nsam - 1) + 1)).sum()
    print(f"Means from fwdpy11:\n{results.mean()}")

    msprime_seeds = []

    for i in range(args.nreps):
        candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        while candidate in msprime_seeds:
            candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        msprime_seeds.append(candidate)
    msprime_results = []
    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = {
            executor.submit(
                run_msprime_get_stats,
                args.N,
                args.nsam,
                args.rho,
                args.theta,
                args.model,
                i,
            )
            for i in msprime_seeds
        }
        for future in concurrent.futures.as_completed(futures):
            msprime_results.append(future.result())
    msprime_results = pd.DataFrame(msprime_results, columns=SimOutput._fields)
    msprime_results["watterson_n"] = (
        msprime_results.Sn / (1.0 / (np.arange(args.nsam - 1) + 1)).sum()
    )
    print(f"Means from msprime:\n{msprime_results.mean()}")



"""
This module provides functions to generate demographic events for
"isolation-with-migration", or IM, models.
"""
import attr
import numpy as np

from fwdpy11.class_decorators import (attr_add_asblack, attr_class_pickle,
                                      attr_class_to_from_dict)

_common_attr_attribs = {
    "frozen": True,
    "auto_attribs": True,
    "repr_ns": "fwdpy11.demographic_models.IM",
}


@attr_add_asblack
# @attr_class_pickle
@attr_class_to_from_dict
@attr.s(**_common_attr_attribs, eq=False)
class TwoDemeIMParameters(object):
    """
    Holds the parameters to :func:`fwdpy11.demographic_models.IM.two_deme_IM`.

    Instances of this class are held as the ``parameters`` attribute of
    :class:`fwdpy11.demographic_models.DemographicModelDetails`.

    Attribute names are the same as the ``kwargs`` to
    :func:`fwdpy11.demographic_models.IM.two_deme_IM`:

    :param Nanc:
    :param T:
    :param psplit:
    :param Ns:
    :param migrates:
    :param burnin:

    .. versionadded:: 0.8.0
    """

    Nanc: int
    T: float
    psplit: float
    Ns: tuple
    migrates: list
    burnin: float

    def __eq__(self, other):
        return all(
            [self.Nanc, self.T, self.psplit, self.Ns, self.burnin]
            == [other.Nanc, other.T, other.psplit, other.Ns, other.burnin]
        ) is True and np.array_equal(self.migrates, other.migrates)


@attr_add_asblack
@attr_class_pickle
@attr_class_to_from_dict
@attr.s(**_common_attr_attribs)
class TwoDemeIMMetaData(object):
    """
    Holds metadata returned by :func:`fwdpy11.demographic_models.IM.two_deme_IM`.

    Instances of this class are held as the ``metadata`` attribute of
    :class:`fwdpy11.demographic_models.DemographicModelDetails`.

    :param split_time: The time until the two demes split
    :param gens_post_split: The time from the split until the end of the simulation
    :param simlen: ``split_time`` + ``gens_post_split``

    .. versionadded:: 0.8.0

    """

    split_time: int
    gens_post_split: int
    simlen: int


def two_deme_IM(Nanc, T, psplit, Ns, migrates, burnin=10.0):
    """
    Isolation-with-migration (IM) model for two demes.

    An ancestral population splits into two daughter demes.
    At the time of the split, ``psplit`` of the ancestral
    population moves into deme 1.  The two daughter populations
    begin exponential growth until the present time and migration
    may occur between them.

    :param Nanc: The ancestral population size.
    :type Nanc: int
    :param T: The time of the split, in units of Nanc generations
              into the past.
    :type T: float
    :param psplit: The proportion of the ancestral population that splits
                   off to found deme 1
    :type psplit: float
    :param Ns: The final sizes of demes 0 and 1, relative to Nanc
    :type Ns: tuple
    :param migrates: The migration rates from 0 to 1 and from 1 to 0,
                     respectively. Migration rates are the fraction
                     of the destination deme replaced by the source
                     deme.
    :type migrates: float
    :param burnin: Time to simulate before the split, in units of Nanc
    :type burnin: float
    :returns: The model events, instances of
              :class:`fwdpy11.demographic_models.IM.TwoDemeIMParameters`
              and :class:`fwdpy11.demographic_models.IM.TwoDemeIMMetaData`.
    :rtype: fwdpy11.demographic_models.DemographicModelDetails

    .. note::

        The events returned by this model assume/require that you will
        construct a population with intitial size ``Nanc``.

    .. versionadded:: 0.6.0

    .. versionchanged:: 0.8.0

        Returns instance of :class:`fwdpy11.demographic_models.DemographicModelDetails`
    """
    import fwdpy11
    import numpy as np
    from .demographic_model_details import DemographicModelDetails

    N0, N1 = Ns
    m01, m10 = migrates

    split_time = np.rint(Nanc * burnin).astype(int)
    # The split event
    split = [
        fwdpy11.move_individuals(
            when=split_time, source=0, destination=1, fraction=psplit
        )
    ]
    # Get growth rates and set growth rate changes,
    # taking care to handle our rounding!
    gens_post_split = np.rint(Nanc * T).astype(int)
    N0split = np.rint(Nanc * (1.0 - psplit))
    if N0split == 0 or N0split == Nanc:
        raise ValueError("invalid value for psplit: {}".format(psplit))
    N0final = np.rint(N0 * Nanc)
    N1split = np.rint(Nanc * psplit)
    if N1split == 0 or N1split == Nanc:
        raise ValueError("invalid value for psplit: {}".format(psplit))
    N1final = np.rint(N1 * Nanc)
    G0 = fwdpy11.exponential_growth_rate(N0split, N0final, gens_post_split)
    G1 = fwdpy11.exponential_growth_rate(N1split, N1final, gens_post_split)
    growth = [
        fwdpy11.SetExponentialGrowth(split_time, 0, G0),
        fwdpy11.SetExponentialGrowth(split_time, 1, G1),
    ]

    # Set up the migration matrix for two demes, but only
    # deme zero exists.
    m = fwdpy11.migration_matrix_single_extant_deme(2, 0)
    # The rows of the matrix change at the split:
    cm = [
        fwdpy11.SetMigrationRates(split_time, 0, [1.0 - m10, m10]),
        fwdpy11.SetMigrationRates(split_time, 1, [m01, 1.0 - m01]),
    ]

    mdict = {
        "mass_migrations": split,
        "set_growth_rates": growth,
        "set_migration_rates": cm,
        "migmatrix": m,
    }

    return DemographicModelDetails(
        model=fwdpy11.DiscreteDemography(**mdict),
        name="Two deme isolation-with-migration (IM) model",
        source={"function": "fwdpy11.demographic_models.IM.two_deme_IM"},
        parameters=TwoDemeIMParameters(
            Nanc=Nanc, T=T, psplit=psplit, Ns=Ns, migrates=migrates, burnin=burnin,
        ),
        citation=None,
        metadata=TwoDemeIMMetaData(
            split_time=split_time,
            gens_post_split=gens_post_split,
            simlen=split_time + gens_post_split,
        ),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# fwdpy11 documentation build configuration file, created by
# sphinx-quickstart on Mon Mar 20 12:22:19 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import pkg_resources


# import subprocess

if (os.environ.get("READTHEDOCS") == "True") is False:
    try:
        sys.path.insert(0, os.path.abspath(".."))
        import fwdpy11  # NOQA
    except:  # NOQA
        raise
else:
    import site

    p = site.getsitepackages()[0]
    sys.path.insert(0, p)

nitpicky = True
nitpick_ignore = [
    ("py:class", "fwdpy11._fwdpy11.GSLrng"),
    ("py:class", "fwdpy11._fwdpy11.PopulationBase"),
    ("py:class", "fwdpy11._fwdpy11.TreeIterator"),
    ("py:class", "fwdpy11._fwdpy11.DiploidPopulation"),
    ("py:class", "fwdpy11._fwdpy11.TableCollection"),
    ("py:class", "fwdpy11._fwdpy11.NodeTable"),
    ("py:class", "fwdpy11._fwdpy11.SampleRecorder"),
    ("py:class", "fwdpy11._fwdpy11.RandomAncientSamples"),
    ("py:class", "fwdpy11._fwdpy11.GeneticValueNoise"),
    ("py:class", "fwdpy11._fwdpy11.GeneticValueIsTrait"),
    ("py:class", "fwdpy11.discrete_demography.MassMigration"),
    ("py:class", "fwdpy11.genetic_values.Optimum"),
    ("py:class", "fwdpy11.genetic_values.PleiotropicOptima"),
    ("py:class", "numpy.uint16"),
    ("py:mod", "msprime"),
    ("py:mod", "tskit"),
    # ("py:class", "pybind11_builtins.pybind11_object"),
    # ("py:class", "fwdpy11._regions.Region"),
    # ("py:class", "fwdpy11._regions.Sregion"),
    # ("py:class", "object"),
    # ("py:obj", "tuple"),
]
#    subprocess.Popen(['doxygen', 'fwdpp.doxygen'],env=my_env)
# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "sphinx_issues",
]

issues_github_path = "molpopgen/fwdpy11"

intersphinx_mapping = {
    "https://docs.python.org/3/": None,
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "sparse": ("https://sparse.pydata.org/en/latest/", None),
    "tskit": ("https://tskit.readthedocs.io/en/stable", None),
    "msprime": ("https://msprime.readthedocs.io/en/stable", None),
    "attrs": ("https://www.attrs.org/en/stable", None),
}

# 'breathe']

# if (os.environ.get('READTHEDOCS')=="True") is False:
#    my_env=os.environ.copy()
#    my_env["LD_LIBRARY_PATH"]=os.path.join(sys.prefix,'lib')
#    subprocess.Popen(['doxygen', 'fwdpy11.doxygen'],env=my_env)
#    subprocess.Popen(['doxygen', 'fwdpp.doxygen'],env=my_env)
#    breathe_projects={'fwdpy11':'xml','fwdpp':'fwdpp_xml'}
#    breathe_default_project="fwdpy11"
# Add any paths that contain templates here, relative to this directory.
templates_path = ["fwdpy11_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "fwdpy11"
copyright = "2017, Kevin Thornton"
author = "Kevin Thornton"

# http://stackoverflow.com/questions/5599254/how-to-use-sphinxs-autodoc-to-document-a-classs-init-self-method


def skip(app, what, name, obj, skip, options):
    if name == "__init__":
        return False
    return skip


def setup(app):
    app.connect("autodoc-skip-member", skip)


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
try:
    from setuptools_scm import get_version

    release = get_version(root="..", relative_to=__file__)
    version = release[:3]
except pkg_resources.DistributionNotFound:
    release = "0.0.0"
    version = "0.0.0"
# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "unused_pages"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "default"

if (os.environ.get("READTHEDOCS") == "True") is True:
    html_theme_options = {
        "github_user": "molpopgen",
        "github_repo": "fwdpy11",
        #        'github_button':True,
        #        'github_banner':True,
    }

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "fwdpy11doc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "fwdpy11.tex", "fwdpy11 Documentation", "Kevin Thornton", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "fwdpy11", "fwdpy11 Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "fwdpy11",
        "fwdpy11 Documentation",
        author,
        "fwdpy11",
        "One line description of project.",
        "Miscellaneous",
    ),
]



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

""" fwdpy11

    isort:skip_file
"""
import sys

from fwdpy11._version import version as __version__  # NOQA

from ._demography import *  # NOQA
from ._dev import *  # NOQA
from ._fwdpy11 import *  # NOQA
from ._monkeypatch import _data_matrix  # NOQA
from ._monkeypatch import _diploid_population  # NOQA
from ._monkeypatch import _table_collection  # NOQA
from .discrete_demography import (  # NOQA
    DiscreteDemography,
    MassMigration,
    move_individuals,
    copy_individuals,
    SetDemeSize,
    MigrationMatrix,
    SetExponentialGrowth,
    SetMigrationRates,
    SetSelfingRate,
)
from .regions import *  # NOQA
from .genetic_map_unit import (
    PoissonPoint,
    PoissonInterval,
    BinomialInterval,
    BinomialPoint,
    FixedCrossovers,
)
from .genetic_values import (  # NOQA
    PleiotropicOptima,
    Optimum,
    GSS,
    GSSmo,
    MultivariateGSS,
    MultivariateGSSmo,
    NoNoise,
    GaussianNoise,
    Additive,
    Multiplicative,
    GBR,
    StrictAdditiveMultivariateEffects,
)
from ._types.demography_debugger import DemographyDebugger  # NOQA
from ._types.model_params import ModelParams, MutationAndRecombinationRates  # NOQA
from ._evolve_genomes import *  # NOQA
from ._evolvets import *  # NOQA

if sys.version_info[0] < 3:
    raise ValueError("Python3 required!")


# NOTE: some operations that can be implemented efficiently
# in Python are supplied as monkey-patches to the pybind11 classes
_monkeypatch._diploid_population._patch_diploid_population(DiploidPopulation)  # NOQA
_monkeypatch._table_collection._patch_table_collection(TableCollection)  # NOQA
_monkeypatch._data_matrix._patch_data_matrix(DataMatrix)  # NOQA



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# This file contains round-trip tests of correctness using
# fwdpy11.mvDES to specify different DES in different demes.
# Low-level tests of fwdpy11.mvDES itself are in test_regions.py

import unittest

import numpy as np

import fwdpy11


def gvalue_multiplicative(pop, ind, scaling):
    g = 1.0
    keys = [k for k in pop.haploid_genomes[pop.diploids[ind].first].smutations]
    keys.extend([k for k in pop.haploid_genomes[pop.diploids[ind].second].smutations])
    keys = np.array(keys, dtype=np.uint32)
    key_counts = np.unique(keys, return_counts=True)

    ind_md = pop.diploid_metadata[ind]
    for i, j in zip(key_counts[0], key_counts[1]):
        if j == 1:
            g *= (
                1.0
                + pop.mutations[i].esizes[ind_md.deme]
                * pop.mutations[i].heffects[ind_md.deme]
            )
        elif j == 2:
            g *= 1.0 + scaling * pop.mutations[i].esizes[ind_md.deme]

    return g


class TestMassMigrationsWithCopies(unittest.TestCase):
    """
    Mass migration happens via copies in generation 1.
    This class uses a custom recorder to detect those copied
    individuals and make sure that their fitnesses are all
    <= 1 because the DES has all mutations being harmful
    in that deme.
    """

    @classmethod
    def setUpClass(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1)
        vcv_matrix = np.array([1, 0.99, 0.99, 1.0]).reshape((2, 2))
        mvDES = fwdpy11.mvDES(
            [fwdpy11.ConstantS(0, 1, 1, 0.1), fwdpy11.ConstantS(0, 1, 1, -0.1)],
            np.zeros(2),
            vcv_matrix,
        )
        copies = [fwdpy11.copy_individuals(1, 0, 1, 1.0)]

        self.pdict = {
            "nregions": [],
            "sregions": [mvDES],
            "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
            "rates": (0, 1, None),
            "gvalue": fwdpy11.Multiplicative(2.0, ndemes=2),
            "demography": fwdpy11.DiscreteDemography(mass_migrations=copies),
            "simlen": 2,
            "prune_selected": True,
        }
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(918273)

        class CheckFitnesses(object):
            """
            Track the fitnesses of individuals who are
            "mass copied"
            """

            def __init__(self):
                self.data = []

            def __call__(self, pop, sampler):
                for i in range(len(pop.diploids), len(pop.diploid_metadata)):
                    md = pop.diploid_metadata[i]
                    dip = pop.diploids[md.label]
                    a = len(pop.haploid_genomes[dip.first].smutations)
                    b = len(pop.haploid_genomes[dip.second].smutations)
                    self.data.append(
                        (
                            pop.generation,
                            pop.diploid_metadata[i].w,
                            pop.diploid_metadata[i].deme,
                            a + b,
                        )
                    )

        self.f = CheckFitnesses()
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100, self.f)
        assert len(self.f.data) > 0, "No data recorded so test is useless"
        assert any(
            [i[3] > 0 for i in self.f.data]
        ), "No mutations in copied individuals so test is useless"

    def test_genetic_values(self):
        for i in self.f.data:
            self.assertEqual(i[2], 1)  # deme == 1
            self.assertTrue(i[1] <= 1.0)  # Mutations are harmful in this deme
            self.assertEqual(i[0], 1)  # Generation == 1, when mass mig happened


class TestMassMigrationsWithMoves(unittest.TestCase):
    """
    Mass migration happens via moves in generation 1.
    Half the individuals are moved from deme 0 to found deme 1
    This class uses a custom recorder to detect those copied
    individuals and make sure that their fitnesses are all
    <= 1 because the DES has all mutations being harmful
    in that deme.
    """

    @classmethod
    def setUpClass(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1)
        vcv_matrix = np.array([1, 0.99, 0.99, 1.0]).reshape((2, 2))
        mvDES = fwdpy11.mvDES(
            [fwdpy11.ConstantS(0, 1, 1, 0.1), fwdpy11.ConstantS(0, 1, 1, -0.1)],
            np.zeros(2),
            vcv_matrix,
        )
        moves = [fwdpy11.move_individuals(1, 0, 1, 0.5)]

        self.pdict = {
            "nregions": [],
            "sregions": [mvDES],
            "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
            "rates": (0, 1, None),
            "gvalue": fwdpy11.Multiplicative(2.0, ndemes=2),
            "demography": fwdpy11.DiscreteDemography(mass_migrations=moves),
            "simlen": 2,
            "prune_selected": True,
        }
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(918273)

        class CheckFitnesses(object):
            """
            Track the fitnesses of individuals who are
            "mass moved"
            """

            def __init__(self):
                self.data = []

            def __call__(self, pop, sampler):
                # Moves don't affect the amount of metadata
                # vs diploid genotypes
                assert len(pop.diploids) == len(pop.diploid_metadata)
                if pop.generation == 1:
                    for i in range(len(pop.diploids)):
                        md = pop.diploid_metadata[i]
                        dip = pop.diploids[md.label]
                        a = len(pop.haploid_genomes[dip.first].smutations)
                        b = len(pop.haploid_genomes[dip.second].smutations)
                        self.data.append(
                            (
                                pop.generation,
                                pop.diploid_metadata[i].w,
                                pop.diploid_metadata[i].deme,
                                a + b,
                            )
                        )

        self.f = CheckFitnesses()
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100, self.f)
        assert len(self.f.data) > 0, "No data recorded so test is useless"
        assert any(
            [i[3] > 0 for i in self.f.data]
        ), "No mutations in copied individuals so test is useless"

    def test_recorded_data(self):
        deme_counts = np.unique(
            np.array([i[2] for i in self.f.data]), return_counts=True
        )
        self.assertEqual(len(deme_counts[0]), 2)
        for i in [0, 1]:
            self.assertEqual(deme_counts[1][i], self.pop.N // 2)
        for i in self.f.data:
            if i[2] == 0:  # deme 0
                self.assertTrue(i[1] >= 1.0)  # Mutations are beneficial in this deme
                self.assertEqual(i[0], 1)  # Generation == 1, when mass mig happened
            else:
                self.assertEqual(i[2], 1)
                self.assertTrue(i[1] <= 1.0)  # Mutations are harmful in this deme


class TestMultiplicativeWithExpSNoMigration(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation([50, 50], 1)
        mvDES = fwdpy11.mvDES(
            [fwdpy11.ExpS(0, 1, 1, 0.1), fwdpy11.ExpS(0, 1, 1, -0.1)],
            np.zeros(2),
            np.identity(2),
        )

        self.pdict = {
            "nregions": [],
            "sregions": [mvDES],
            "recregions": [fwdpy11.PoissonInterval(0, 1, 1e-2)],
            "rates": (0, 1e-2, None),
            "gvalue": fwdpy11.Multiplicative(2.0, ndemes=2),
            "demography": fwdpy11.DiscreteDemography(),
            "simlen": 100,
            "prune_selected": True,
        }
        self.params = fwdpy11.ModelParams(**self.pdict)
        self.rng = fwdpy11.GSLrng(918273)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        n0 = 0
        n1 = 0
        for i, md in enumerate(self.pop.diploid_metadata):
            ng0 = len(self.pop.haploid_genomes[self.pop.diploids[i].first].smutations)
            ng1 = len(self.pop.haploid_genomes[self.pop.diploids[i].second].smutations)
            if ng0 + ng1 > 0:
                if md.deme == 0:
                    n0 += 1
                elif md.deme == 1:
                    n1 += 1
        if n0 == 0 or n1 == 0:
            self.assertFail("we don't have individuals with mutations in each deme")

    def test_it(self):
        gv = np.zeros(self.pop.N)
        for i in range(self.pop.N):
            gv[i] = gvalue_multiplicative(self.pop, i, 2.0)
        md = np.array(self.pop.diploid_metadata, copy=False)
        self.assertTrue(np.allclose(gv, md["g"]))


class TestGaussianStabilizingSelection(unittest.TestCase):
    """
    Quick sim w/high migration rate
    to put the same mutation in both demes.
    """

    @classmethod
    def setUpClass(self):
        pdict = {
            "nregions": [],
            "sregions": [
                fwdpy11.mvDES(
                    fwdpy11.MultivariateGaussianEffects(
                        0, 1, 1, h=1, cov_matrix=np.identity(2)
                    ),
                    np.zeros(2),
                )
            ],
            "recregions": [],
            "rates": (0, 1e-3, None),
            "demography": fwdpy11.DiscreteDemography(
                migmatrix=np.array([0.9, 0.1, 0.1, 0.9]).reshape((2, 2))
            ),
            "simlen": 100,
            "gvalue": fwdpy11.Additive(
                ndemes=2, scaling=2, gvalue_to_fitness=fwdpy11.GSS(optimum=0.0, VS=1.0)
            ),
        }

        self.params = fwdpy11.ModelParams(**pdict)
        self.pop = fwdpy11.DiploidPopulation([100, 100], 1.0)
        self.rng = fwdpy11.GSLrng(1010)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)

        tv = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        nt = np.array(self.pop.tables.nodes, copy=False)
        demes_with_muts = np.zeros(2)
        for t in tv:
            for m in t.mutations():
                demes = np.unique(nt["deme"][t.samples_below(m.node)])
                for d in demes:
                    demes_with_muts[d] += 1
        assert np.all(demes_with_muts > 0), "test requires mutations in both demes"

    def test_genetic_values(self):
        for m in self.pop.diploid_metadata:
            g = 0.0
            for i in [
                self.pop.diploids[m.label].first,
                self.pop.diploids[m.label].second,
            ]:
                for k in self.pop.haploid_genomes[i].smutations:
                    g += self.pop.mutations[k].esizes[m.deme]
            self.assertAlmostEqual(m.g, g)


class TestMultivariateLogNormalS(unittest.TestCase):
    """
    Quick sim w/high migration rate
    to put the same mutation in both demes.
    """

    @classmethod
    def setUpClass(self):
        ln = fwdpy11.LogNormalS.mv(0, 1, 1, scaling=-200)
        pdict = {
            "nregions": [],
            "sregions": [fwdpy11.mvDES(ln, np.zeros(2), np.identity(2))],
            "recregions": [],
            "rates": (0, 1e-3, None),
            "demography": fwdpy11.DiscreteDemography(
                migmatrix=np.array([0.9, 0.1, 0.1, 0.9]).reshape((2, 2))
            ),
            "simlen": 100,
            "gvalue": fwdpy11.Multiplicative(ndemes=2, scaling=2),
        }

        self.params = fwdpy11.ModelParams(**pdict)
        self.pop = fwdpy11.DiploidPopulation([100, 100], 1.0)
        self.rng = fwdpy11.GSLrng(201012)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)

        tv = fwdpy11.TreeIterator(
            self.pop.tables, self.pop.alive_nodes, update_samples=True
        )
        nt = np.array(self.pop.tables.nodes, copy=False)
        demes_with_muts = np.zeros(2)
        for t in tv:
            for m in t.mutations():
                demes = np.unique(nt["deme"][t.samples_below(m.node)])
                for d in demes:
                    demes_with_muts[d] += 1
        assert np.all(demes_with_muts > 0), "test requires mutations in both demes"

    def test_genetic_values(self):
        for i, m in enumerate(self.pop.diploid_metadata):
            g = gvalue_multiplicative(self.pop, i, 2.0)
            self.assertAlmostEqual(m.g, g)


if __name__ == "__main__":
    unittest.main()
